// Copyright Yangqing Jia 2013

#ifndef CAFFE_VISUALIZER_HPP_
#define CAFFE_VISUALIZER_HPP_

#include <string>
#include <vector>

namespace caffe {

template <typename Dtype>
class Visualizer {
 public:
  explicit Visualizer(const SolverParameter& param);
  // The main entry of the visualizer function.
  void Visualize(const char* net_file = NULL, const int display_node = -1);
  void Visualize(const string net_file, const int display_node = -1) { Visualize(net_file.c_str(), display_node); }
  ~Visualizer() {}
  inline Net<Dtype>* net() { return net_.get(); }

 protected:
  // PreVisualize is run before any visualization iteration starts, allowing one to
  // put up some scaffold.
  void PreVisualize(int display_node);
  void PreLeapFrog();
  // Get the update value for the current iteration.
  void ComputeUpdateValue();
  // The Solver::Snapshot function implements the basic snapshotting utility
  // that stores the learned net. You should implement the SnapshotSolverState()
  // function that produces a SolverState protocol buffer that needs to be
  // written to disk together with the learned net.
  void Snapshot();
  // The Restore function implements how one loads the net to visualize.
  void Restore(const char* net_file);
  // The Monte Carlo routine
  void MonteCarlo();
  Dtype GetLearningRate();



  SolverParameter param_;
  int iter_, step_;
  shared_ptr<Net<Dtype> > net_;

  // samples
  Blob<Dtype> sample_;
  // labels
  Blob<Dtype> label_;
  int display_;
  // history
  Blob<Dtype> history_sample_;
  Blob<Dtype> history_momentum_;
  // momentum term
  Blob<Dtype> momentum_;
  Dtype T_;

  DISABLE_COPY_AND_ASSIGN(Visualizer);
};

}  // namespace caffe

#endif  // CAFFE_OPTIMIZATION_SOLVER_HPP_
