// Copyright Yangqing Jia 2013

#include <cstdio>

#include <algorithm>
#include <string>
#include <vector>

#include "caffe/net.hpp"
#include "caffe/proto/caffe.pb.h"
#include "caffe/visualizer.hpp"
#include "caffe/util/io.hpp"
#include "caffe/util/math_functions.hpp"
#include "caffe/filler.hpp"
#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"

using std::max;
using std::min;

namespace caffe {

template <typename Dtype>
Visualizer<Dtype>::Visualizer(const SolverParameter& param)
    : param_(param), net_() {
	// Scaffolding code
	NetParameter train_net_param;
	ReadProtoFromTextFile(param_.train_net(), &train_net_param);
	LOG(INFO) << "Creating net.";
	net_.reset(new Net<Dtype>(train_net_param));
	LOG(INFO) << "Visualizer scaffolding done.";

    // set the input blobs
	CHECK_EQ(train_net_param.input_size(), 2)
		      << "Incorrect bottom blob specifications. It should be sample and label";
	CHECK_EQ(train_net_param.input_size() * 4, train_net_param.input_dim_size())
	      << "Incorrect bottom blob dimension specifications.";
	sample_.Reshape(train_net_param.input_dim(0 * 4),
			train_net_param.input_dim(0 * 4 + 1),
			train_net_param.input_dim(0 * 4 + 2),
			train_net_param.input_dim(0 * 4 + 3));
	history_sample_.Reshape(train_net_param.input_dim(0 * 4),
			train_net_param.input_dim(0 * 4 + 1),
			train_net_param.input_dim(0 * 4 + 2),
			train_net_param.input_dim(0 * 4 + 3));

	label_.Reshape(train_net_param.input_dim(1 * 4),
				train_net_param.input_dim(1 * 4 + 1),
				train_net_param.input_dim(1 * 4 + 2),
				train_net_param.input_dim(1 * 4 + 3));

	//set the momentum
	momentum_.Reshape(train_net_param.input_dim(0 * 4),
			train_net_param.input_dim(0 * 4 + 1),
			train_net_param.input_dim(0 * 4 + 2),
			train_net_param.input_dim(0 * 4 + 3));
	history_momentum_.Reshape(train_net_param.input_dim(0 * 4),
			train_net_param.input_dim(0 * 4 + 1),
			train_net_param.input_dim(0 * 4 + 2),
			train_net_param.input_dim(0 * 4 + 3));
}


template <typename Dtype>
void Visualizer<Dtype>::Visualize(const char* net_file, int display_node) {
  Caffe::set_mode(Caffe::Brew(param_.solver_mode()));
  if (param_.solver_mode() && param_.has_device_id()) {
    Caffe::SetDevice(param_.device_id());
  }
  Caffe::set_phase(Caffe::TRAIN);
  LOG(INFO) << "Visualizing " << net_->name();
  PreVisualize(display_node);


  if (net_file) {
    LOG(INFO) << "Loading model from " << net_file;
    Restore(net_file);
  }

  iter_ = 0;
  step_ = 0;
  //Snapshot();
  T_ = 1;// param_.max_iter();
  Dtype sigma = (Dtype)this->param_.sigma();

  //bool ischange = false;
  for(iter_ =  1; iter_ <= param_.max_iter(); iter_++) {
	LOG(INFO) << "Iter: " << iter_;
	LOG(INFO) << "Tempeature is " << T_;	
	PreLeapFrog();
	Dtype loss = 0.0;
	Dtype data_rate = GetLearningRate() / (Dtype)this->param_.gamma();
	Dtype momentum_rate = GetLearningRate();
	Dtype regularizer = (Dtype)this->param_.power();

	// Compute derative for equation (11)
	vector<Blob<Dtype>*> bottom_vec;
	bottom_vec.push_back(&sample_);
	bottom_vec.push_back(&label_);
	loss = net_->ForwardBackward(bottom_vec);

/*	for(step_ = 1; step_  <= param_.stepsize(); step_++ )
	{
		// equation (11)
		caffe_axpby(momentum_.count(), -momentum_rate / Dtype(1.0) * regularizer,
			net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
			momentum_.mutable_cpu_data());

		// equation (12)
		caffe_axpby(momentum_.count(), data_rate,
			momentum_.cpu_data(), Dtype(1.0),
			momentum_.mutable_cpu_data());

		// update derative for equation (13)
		bottom_vec.clear();
		bottom_vec.push_back(&sample_);
		bottom_vec.push_back(&label_);
		loss = net_->ForwardBackward(bottom_vec);

		// equation (13)
		caffe_axpby(momentum_.count(), -momentum_rate / Dtype(1.0) * regularizer,
			net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
			momentum_.mutable_cpu_data());
	}*/
	caffe_axpby(momentum_.count(), -momentum_rate/Dtype(1.0)*regularizer,
		net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
		momentum_.mutable_cpu_data());
	caffe_axpby(momentum_.count(), -momentum_rate/Dtype(1.0)/sigma,
		sample_.cpu_data(), Dtype(1.0),
		momentum_.mutable_cpu_data());

	/*int test_id = 0;
	const Dtype* net_diff = net_->input_blobs()[0]->cpu_diff();
	for(int ii = 0; ii < momentum_.count(); ii++)
	{
		if(net_diff[ii]!= 0)
			test_id = 1;
	}
	LOG(INFO) << "TEST id is " << test_id;*/

	for(step_ =  1; step_ <= param_.stepsize(); step_++) {
		// Perform leap-frog algorithm
		

		//for (int i=0; i<5; i++) {
		//	LOG(INFO) << sample_.cpu_data()[i] << net_->input_blobs()[0]->cpu_diff()[i];
		//} 
		//LOG(INFO) << "ComputeUpdateValue";
		/*bottom_vec.clear();
		bottom_vec.push_back(&sample_);
		bottom_vec.push_back(&label_);
		loss = net_->ForwardBackward(bottom_vec);

		caffe_axpby(momentum_.count(), -momentum_rate,
			net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
			momentum_.mutable_cpu_data());*/
		
		caffe_axpby(sample_.count(), data_rate,
			momentum_.cpu_data(), Dtype(1.0),
			sample_.mutable_cpu_data());

		bottom_vec.clear();
		bottom_vec.push_back(&sample_);
		bottom_vec.push_back(&label_);
		loss = net_->ForwardBackward(bottom_vec);

		if(step_ < param_.stepsize()) {

			caffe_axpby(momentum_.count(), -momentum_rate * regularizer,
				net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
				momentum_.mutable_cpu_data());
			caffe_axpby(momentum_.count(), -momentum_rate/sigma,
				sample_.cpu_data(), Dtype(1.0),
				momentum_.mutable_cpu_data());
		} else {
			caffe_axpby(momentum_.count(), -momentum_rate/Dtype(2.0)*regularizer,
				net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
				momentum_.mutable_cpu_data());
			caffe_axpby(momentum_.count(), -momentum_rate/Dtype(2.0)/sigma,
				sample_.cpu_data(), Dtype(1.0),
				momentum_.mutable_cpu_data());		
		}


		// Snapshot();
	}
	LOG(INFO) << "loss: " << loss;

	// Determine whether to make the leap-frog move
	MonteCarlo();

	//T_ += 20;
	//T_ = min(T_, (Dtype)1000.0);

	if (param_.snapshot() && iter_ % param_.snapshot() == 0) {
		Snapshot();
	}
  }

  LOG(INFO) << "Visualization Done.";
}

template <typename Dtype>
void Visualizer<Dtype>::MonteCarlo() {
	// Evaluate potential and kinetic energies
	int num = sample_.num();
	int dim = sample_.count() / sample_.num();

	vector<Blob<Dtype>*> history_bottom_vec;
	history_bottom_vec.push_back(&history_sample_);
	history_bottom_vec.push_back(&label_);
	const vector<Blob<Dtype>*>& temp_history_result = net_->Forward(history_bottom_vec);
	Blob<Dtype> history_result(temp_history_result[0]->num(),
			temp_history_result[0]->channels(), temp_history_result[0]->height(), temp_history_result[0]->width());
	history_result.CopyFrom(*temp_history_result[0]);


	vector<Blob<Dtype>*> bottom_vec;
	bottom_vec.push_back(&sample_);
	bottom_vec.push_back(&label_);
	const vector<Blob<Dtype>*>& temp_proposed_result = net_->Forward(bottom_vec);
	Blob<Dtype> proposed_result(temp_proposed_result[0]->num(),
			temp_proposed_result[0]->channels(), temp_proposed_result[0]->height(), temp_proposed_result[0]->width());
	proposed_result.CopyFrom(*temp_proposed_result[0]);

	Dtype regularizer = (Dtype)this->param_.power();

	int count_accept = 0;
	Dtype sigma = (Dtype)this->param_.sigma();

	for (int i = 0; i < num; i++) {
		Dtype history_U = 0.0;
		Dtype history_K = 0.0;
		Dtype proposed_U = 0.0;
		Dtype proposed_K = 0.0;

		history_U = regularizer*history_result.cpu_data()[i];
		//LOG(INFO) << "hostory_U: " << history_U;
		const Dtype* history_momentum_data = history_momentum_.cpu_data();
		const Dtype* history_sample_data = history_sample_.cpu_data();
		for (int d = 0; d < dim; d++) {
			history_K += history_momentum_data[i*dim+d]*history_momentum_data[i*dim+d]/2.0;
			history_U += history_sample_data[i*dim+d]*history_sample_data[i*dim+d]/(2.0*sigma*sigma);
		}
		//history_K /= dim;
		history_K /= (2.0*(Dtype)this->param_.gamma());
		proposed_U = regularizer*proposed_result.cpu_data()[i];
		//LOG(INFO) << "proposed_U: " << proposed_U;
		const Dtype* proposed_momentum_data = momentum_.cpu_data();
		const Dtype* proposed_sample_data = sample_.cpu_data();
		for (int d = 0; d < dim; d++) {
			proposed_K += proposed_momentum_data[i*dim+d]*proposed_momentum_data[i*dim+d]/2.0;
			proposed_U += proposed_sample_data[i*dim+d]*proposed_sample_data[i*dim+d]/(2.0*sigma*sigma);
		}
		//proposed_K /= dim;
		proposed_K /= (2.0*(Dtype)this->param_.gamma());

		/*LOG(INFO) << "history_U: " << history_U;
		LOG(INFO) << "proposed_U: " << proposed_U;
		LOG(INFO) << "history_K: " << history_K;
		LOG(INFO) << "proposed_K: " << proposed_K;*/

		Dtype random_num = (Dtype)rand()/(Dtype)RAND_MAX;
		LOG(INFO) << "ACCEPT PROB is " << (history_U + history_K - proposed_U - proposed_K); 
		if (random_num < exp((history_U+history_K-proposed_U-proposed_K)/T_)) {
			//Accept
			count_accept++;
			Dtype* m_history_sample_data = history_sample_.mutable_cpu_data();
			Dtype* m_history_momentum_data = history_momentum_.mutable_cpu_data();
			const Dtype* proposed_sample_data = sample_.cpu_data();
			for (int d = 0; d < dim; d++) {
				m_history_sample_data[i*dim+d] = proposed_sample_data[i*dim+d];
				m_history_momentum_data[i*dim+d] = proposed_momentum_data[i*dim+d];
			}
		} else {
			//Reject
			Dtype* m_proposed_sample_data = sample_.mutable_cpu_data();
			Dtype* m_proposed_momentum_data = momentum_.mutable_cpu_data();
			const Dtype* history_sample_data = history_sample_.cpu_data();
			for (int d = 0; d < dim; d++) {
				m_proposed_sample_data[i*dim+d] = history_sample_data[i*dim+d];
				m_proposed_momentum_data[i*dim+d] = history_momentum_data[i*dim+d];
			}
		}
	}

	LOG(INFO) << "Accept rate: " << count_accept/(Dtype)num;
}


template <typename Dtype>
void Visualizer<Dtype>::Snapshot() {
	const Dtype* sample_data = sample_.cpu_data();
	int num = sample_.num();
	int channels = sample_.channels();
	int width = sample_.width();
	int height = sample_.height();
	cv::Mat cv_img(height, width, CV_8UC(3));

	if ((channels!=1) && (channels!=3)) {
		LOG(FATAL) << "Visualizer currently only support data channel of 1 or 3";
	}

	int is_rgb = (channels==3);

	for (int itemid = 0; itemid < num; itemid++) {

		// save sample image
		if (is_rgb) {
			Dtype max_val = -FLT_MAX;
			Dtype min_val = FLT_MAX;
			for (int h = 0; h < height; ++h) {
				for (int w = 0; w < width; ++w) {
					for (int c = 0; c < 3; c++) {
						max_val = std::max(max_val, sample_data[((itemid * channels + c) * height + h) * width + w]);
						min_val = std::min(min_val, sample_data[((itemid * channels + c) * height + h) * width + w]);
					}
				}
			}
			LOG(INFO) << "max_val: "<< max_val << "min_val: "<<min_val;
			Dtype scale_val = 255.0/(max_val-min_val);
			for (int h = 0; h < height; ++h) {
				for (int w = 0; w < width; ++w) {
					for (int c = 0; c < 3; c++) {
						cv_img.at<cv::Vec3b>(h, w)[c] = (unsigned char)(scale_val * (sample_data[((itemid * channels + c) * height + h) * width + w]-min_val));
					}
				}
			}
		/*	for(int c = 0; c < 3; c++) {
			for (int h = 0; h < height; ++h) {
                                for (int w = 0; w < width; ++w) {
                                                max_val = std::max(max_val, sample_data[((itemid * channels + c) * height + h) * width + w]);
                                                min_val = std::min(min_val, sample_data[((itemid * channels + c) * height + h) * width + w]);
                                }
                        }
                        LOG(INFO) << "max_val: "<< max_val << "min_val: "<<min_val;
                        Dtype scale_val = 255.0/(max_val-min_val);
                        for (int h = 0; h < height; ++h) {
                                for (int w = 0; w < width; ++w){ 
                                           cv_img.at<cv::Vec3b>(h, w)[c] = (unsigned char)(scale_val * (sample_data[((itemid * channels + c) * height + h) * width + w]-min_val));
                                }
                        }
			}*/

		}
		else {
			Dtype max_val = -FLT_MAX;
			Dtype min_val = FLT_MAX;
			for (int h = 0; h < height; ++h) {
				for (int w = 0; w < width; ++w) {
					max_val = std::max(max_val, sample_data[(itemid * height + h) * width + w]);
					min_val = std::min(min_val, sample_data[(itemid * height + h) * width + w]);
				}
			}
			LOG(INFO) << "max_val: "<< max_val << "min_val: "<<min_val;
 			Dtype scale_val = 255.0/(max_val-min_val);
			for (int h = 0; h < height; ++h) {
				for (int w = 0; w < width; ++w) {
					for (int c = 0; c < 3; c++) {
						cv_img.at<cv::Vec3b>(h, w)[c] = (unsigned char)(scale_val * (sample_data[(itemid * height + h) * width + w]-min_val));
					}
				}
			}
		}

		string filename(param_.snapshot_prefix());
		const int kBufferSize = 100;
		char iter_str_buffer[kBufferSize];
		snprintf(iter_str_buffer, kBufferSize, "display_%d_iter_%d_step_%d_sample_%d.png", display_, iter_, step_, itemid);
		filename += iter_str_buffer;

		imwrite(filename, cv_img);
	}

	LOG(INFO) << "Snapshot is taken.";
}

template <typename Dtype>
void Visualizer<Dtype>::Restore(const char* net_file) {
	NetParameter net_param;
	ReadProtoFromBinaryFile(net_file, &net_param);
	net_->CopyTrainedLayersFrom(net_param);
}


// Return the current learning rate.
template <typename Dtype>
Dtype Visualizer<Dtype>::GetLearningRate() {
  Dtype rate;
  const string& lr_policy = this->param_.lr_policy();
  if (lr_policy == "fixed") {
    rate = this->param_.base_lr();
  } else {
    LOG(FATAL) << "Unknown learning rate policy: " << lr_policy;
  }

  return rate;
}


template <typename Dtype>
void Visualizer<Dtype>::PreVisualize(int display_node) {
	// Initialize bottom_vec_
	Dtype sigma = (Dtype)this->param_.sigma();
	FillerParameter filler_param;
	filler_param.set_mean(0.);
	filler_param.set_std(1);
	filler_param.set_value(0);
	//GaussianFiller<Dtype> filler(filler_param);
	// initialize sample use constant 0
	ConstantFiller<Dtype> filler(filler_param);
	filler.Fill(&sample_);

	//project sample back to be within range [0, 255]
	/* Dtype* sample_data = sample_.mutable_cpu_data();
	for (int i=0; i<sample_.count(); i++) {
		sample_data[i] = std::min(sample_data[i], (Dtype)255.0);
		sample_data[i] = std::max(sample_data[i], (Dtype)0.0);
	} */

	if (display_node < 0) {
		display_ = this->param_.display();
	} else {
		display_ = display_node;
	}

	Dtype* label_data = label_.mutable_cpu_data();
	for (int i = 0; i < label_.count(); i++) {
		label_data[i] = display_;
	}
}


template <typename Dtype>
void Visualizer<Dtype>::PreLeapFrog() {
	// Initialize the momentum term and save the current bottom_vec_[0] and momentum terms
	FillerParameter filler_param;
	filler_param.set_mean(0.);
	filler_param.set_std(sqrt(this->param_.gamma()));
	GaussianFiller<Dtype> filler(filler_param);
	filler.Fill(&momentum_);

	caffe_copy(sample_.count(),
			sample_.cpu_data(),
			history_sample_.mutable_cpu_data());
	caffe_copy(momentum_.count(),
			momentum_.cpu_data(),
			history_momentum_.mutable_cpu_data());
}


template <typename Dtype>
void Visualizer<Dtype>::ComputeUpdateValue() {
	// LOG(INFO) << "step_: " << step_;
	// get the learning rate
	Dtype data_rate = GetLearningRate() / (Dtype)this->param_.gamma();
	Dtype regularizer = (Dtype)this->param_.power();
	if (step_ == 0) {
	  data_rate = 0.;
	}
	Dtype momentum_rate = GetLearningRate();
	if ((step_ == 0) || (step_ == this->param_.stepsize()-1)) {
	  momentum_rate /= 2.0;
	}
	if (step_ == this->param_.stepsize()-2) {
	  momentum_rate = 0.;
	}

	/*for (int i=0; i<5; i++) {
		LOG(INFO) << sample_.cpu_data()[i] << net_->input_blobs()[0]->cpu_diff()[i];
	} 
	LOG(INFO) << "ComputeUpdateValue";*/

	caffe_axpby(momentum_.count(), -momentum_rate*regularizer,
		net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
		momentum_.mutable_cpu_data());
	caffe_axpby(momentum_.count(), -momentum_rate,
		sample_.cpu_data(), Dtype(1.0),
		momentum_.mutable_cpu_data());

	caffe_axpby(sample_.count(), data_rate,
		momentum_.cpu_data(), Dtype(1.0),
		sample_.mutable_cpu_data());

	/* caffe_axpby(sample_.count(), -data_rate,
			net_->input_blobs()[0]->cpu_diff(), Dtype(1.0),
			sample_.mutable_cpu_data()); */


	//project sample back to be within range [0, 255]
	/* Dtype* sample_data = sample_.mutable_cpu_data();
	for (int i=0; i<sample_.count(); i++) {
		sample_data[i] = std::min(sample_data[i], (Dtype)255.0);
		sample_data[i] = std::max(sample_data[i], (Dtype)0.0);
	} */
}


INSTANTIATE_CLASS(Visualizer);

}  // namespace caffe
